# Copyright 2022 Infinidat Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for INFINIDAT InfiniBox volume driver."""

import collections
import copy
import functools
import itertools
import platform
import socket
from unittest import mock
import uuid

import ddt
from oslo_utils import units

from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder import version
from cinder.volume import configuration
from cinder.volume.drivers import infinidat


TEST_LUN = 1
TEST_WWN_1 = '00:11:22:33:44:55:66:77'
TEST_WWN_2 = '11:11:22:33:44:55:66:77'
TEST_IP_ADDRESS1 = '1.1.1.1'
TEST_IP_ADDRESS2 = '2.2.2.2'
TEST_IP_ADDRESS3 = '3.3.3.3'
TEST_IP_ADDRESS4 = '4.4.4.4'
TEST_INITIATOR_IQN = 'iqn.2012-07.org.initiator:01'
TEST_INITIATOR_IQN2 = 'iqn.2012-07.org.initiator:02'
TEST_TARGET_IQN = 'iqn.2012-07.org.target:01'
TEST_ISCSI_TCP_PORT1 = 3261
TEST_ISCSI_TCP_PORT2 = 3262
TEST_ISCSI_NAMESPACE1 = 'netspace1'
TEST_ISCSI_NAMESPACE2 = 'netspace2'
TEST_TARGET_PORTAL1 = '{}:{}'.format(TEST_IP_ADDRESS1, TEST_ISCSI_TCP_PORT1)
TEST_TARGET_PORTAL2 = '{}:{}'.format(TEST_IP_ADDRESS2, TEST_ISCSI_TCP_PORT1)
TEST_TARGET_PORTAL3 = '{}:{}'.format(TEST_IP_ADDRESS3, TEST_ISCSI_TCP_PORT2)
TEST_TARGET_PORTAL4 = '{}:{}'.format(TEST_IP_ADDRESS4, TEST_ISCSI_TCP_PORT2)
TEST_FC_PROTOCOL = 'fc'
TEST_ISCSI_PROTOCOL = 'iscsi'
TEST_VOLUME_SOURCE_NAME = 'test-volume'
TEST_VOLUME_TYPE = 'MASTER'
TEST_VOLUME_SOURCE_ID = 12345
TEST_VOLUME_METADATA = {'cinder_id': fake.VOLUME_ID}
TEST_SNAPSHOT_SOURCE_NAME = 'test-snapshot'
TEST_SNAPSHOT_SOURCE_ID = 67890
TEST_SNAPSHOT_METADATA = {'cinder_id': fake.SNAPSHOT_ID}
TEST_POOL_NAME = 'pool'
TEST_POOL_NAME2 = 'pool2'
TEST_SYSTEM_SERIAL = 123
TEST_SYSTEM_SERIAL2 = 456

test_volume = mock.Mock(id=fake.VOLUME_ID, name_id=fake.VOLUME_ID, size=1,
                        volume_type_id=fake.VOLUME_TYPE_ID, group_id=None,
                        multiattach=False, volume_attachment=None)
test_volume2 = mock.Mock(id=fake.VOLUME2_ID, name_id=fake.VOLUME2_ID, size=1,
                         volume_type_id=None, group_id=None,
                         multiattach=False, volume_attachment=None)
test_volume3 = mock.Mock(id=fake.VOLUME3_ID, name_id=fake.VOLUME3_ID, size=1,
                         volume_type_id=fake.VOLUME_TYPE_ID,
                         group_id=fake.GROUP_ID, multiattach=True,
                         volume_attachment=None)
test_snapshot = mock.Mock(id=fake.SNAPSHOT_ID, volume=test_volume)
test_clone = mock.Mock(id=fake.VOLUME4_ID, name_id=fake.VOLUME4_ID, size=1,
                       volume_type_id=fake.VOLUME_TYPE_ID, group_id=None,
                       multiattach=False, volume_attachment=None)
test_group = mock.Mock(id=fake.GROUP_ID)
test_snapgroup = mock.Mock(id=fake.GROUP_SNAPSHOT_ID, group=test_group)
test_connector = dict(wwpns=[TEST_WWN_1],
                      initiator=TEST_INITIATOR_IQN)
test_connector2 = dict(wwpns=[TEST_WWN_2],
                       initiator=TEST_INITIATOR_IQN2)
test_connector3 = dict(wwpns=None, initiator=None)
test_attachment1 = mock.Mock(connector=test_connector)
test_attachment2 = mock.Mock(connector=test_connector2)
test_attachment3 = mock.Mock(connector=None)


def skip_driver_setup(func):
    @functools.wraps(func)
    def f(*args, **kwargs):
        return func(*args, **kwargs)
    f.__skip_driver_setup = True
    return f


class FakeInfinisdkException(Exception):
    pass


class InfiniboxDriverTestCaseBase(test.TestCase):
    def _test_skips_driver_setup(self):
        test_method_name = self.id().split('.')[-1]
        test_method = getattr(self, test_method_name)
        return getattr(test_method, '__skip_driver_setup', False)

    def setUp(self):
        super(InfiniboxDriverTestCaseBase, self).setUp()

        self.configuration = configuration.Configuration(None)
        self.configuration.append_config_values(infinidat.infinidat_opts)
        self.override_config('san_ip', 'infinibox',
                             configuration.SHARED_CONF_GROUP)
        self.override_config('san_login', 'user',
                             configuration.SHARED_CONF_GROUP)
        self.override_config('san_password', 'password',
                             configuration.SHARED_CONF_GROUP)
        self.override_config('infinidat_pool_name', TEST_POOL_NAME)
        self.driver = infinidat.InfiniboxVolumeDriver(
            configuration=self.configuration)
        self._system = self._infinibox_mock()
        # mock external library dependencies
        infinisdk = self.patch("cinder.volume.drivers.infinidat.infinisdk")
        capacity = self.patch("cinder.volume.drivers.infinidat.capacity")
        self._log = self.patch("cinder.volume.drivers.infinidat.LOG")
        self._iqn = self.patch("cinder.volume.drivers.infinidat.iqn")
        self._wwn = self.patch("cinder.volume.drivers.infinidat.wwn")
        self._wwn.WWN = mock.Mock
        self._iqn.IQN = mock.Mock
        capacity.byte = 1
        capacity.GiB = units.Gi
        infinisdk.core.exceptions.InfiniSDKException = FakeInfinisdkException
        infinisdk.InfiniBox.return_value = self._system

        if not self._test_skips_driver_setup():
            self.driver.do_setup(None)

    def _infinibox_mock(self):
        result = mock.Mock()
        self._mock_volume = mock.Mock()
        self._mock_new_volume = mock.Mock()
        self._mock_volume.get_id.return_value = TEST_VOLUME_SOURCE_ID
        self._mock_volume.get_name.return_value = TEST_VOLUME_SOURCE_NAME
        self._mock_volume.get_type.return_value = TEST_VOLUME_TYPE
        self._mock_volume.get_pool_name.return_value = TEST_POOL_NAME
        self._mock_volume.get_size.return_value = 1 * units.Gi
        self._mock_volume.has_children.return_value = False
        self._mock_volume.get_qos_policy.return_value = None
        self._mock_volume.get_logical_units.return_value = []
        self._mock_volume.get_all_metadata.return_value = {}
        self._mock_volume.create_snapshot.return_value = self._mock_volume
        self._mock_snapshot = mock.Mock()
        self._mock_snapshot.get_parent.return_value = self._mock_volume
        self._mock_host = mock.Mock()
        self._mock_host.get_luns.return_value = []
        self._mock_host.map_volume().get_lun.return_value = TEST_LUN
        self._mock_pool = mock.Mock()
        self._mock_pool.get_free_physical_capacity.return_value = units.Gi
        self._mock_pool.get_physical_capacity.return_value = units.Gi
        self._mock_pool.get_volumes.return_value = [self._mock_volume]
        self._mock_name_space1 = mock.Mock()
        self._mock_name_space2 = mock.Mock()
        self._mock_name_space1.get_ips.return_value = [
            mock.Mock(ip_address=TEST_IP_ADDRESS1, enabled=True)]
        self._mock_name_space2.get_ips.return_value = [
            mock.Mock(ip_address=TEST_IP_ADDRESS3, enabled=True)]
        self._mock_name_space1.get_properties.return_value = mock.Mock(
            iscsi_iqn=TEST_TARGET_IQN, iscsi_tcp_port=TEST_ISCSI_TCP_PORT1)
        self._mock_name_space2.get_properties.return_value = mock.Mock(
            iscsi_iqn=TEST_TARGET_IQN, iscsi_tcp_port=TEST_ISCSI_TCP_PORT2)
        self._mock_group = mock.Mock()
        self._mock_qos_policy = mock.Mock()
        result.volumes.safe_get.return_value = self._mock_volume
        result.volumes.create.return_value = self._mock_volume
        result.pools.safe_get.return_value = self._mock_pool
        result.hosts.safe_get.return_value = self._mock_host
        result.cons_groups.safe_get.return_value = self._mock_group
        result.cons_groups.create.return_value = self._mock_group
        result.hosts.create.return_value = self._mock_host
        result.network_spaces.safe_get.return_value = self._mock_name_space1
        result.components.nodes.get_all.return_value = []
        result.qos_policies.create.return_value = self._mock_qos_policy
        result.qos_policies.safe_get.return_value = None
        result.get_serial.return_value = TEST_SYSTEM_SERIAL
        return result

    def _raise_infinisdk(self, *args, **kwargs):
        raise FakeInfinisdkException()


@ddt.ddt
class InfiniboxDriverTestCase(InfiniboxDriverTestCaseBase):
    def _generate_mock_object_metadata(self, cinder_object):
        return {"system": "openstack",
                "openstack_version": version.version_info.release_string(),
                "cinder_id": cinder_object.id,
                "cinder_name": cinder_object.name,
                "host.created_by": infinidat._INFINIDAT_CINDER_IDENTIFIER}

    def _validate_object_metadata(self, infinidat_object, cinder_object):
        infinidat_object.set_metadata_from_dict.assert_called_once_with(
            self._generate_mock_object_metadata(cinder_object))

    def _generate_mock_host_metadata(self):
        return {"system": "openstack",
                "openstack_version": version.version_info.release_string(),
                "hostname": socket.gethostname(),
                "platform": platform.platform(),
                "host.created_by": infinidat._INFINIDAT_CINDER_IDENTIFIER}

    def _validate_host_metadata(self):
        self._mock_host.set_metadata_from_dict.assert_called_once_with(
            self._generate_mock_host_metadata())

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_get_oslo_driver_opts')
    def test_get_driver_options(self, _get_oslo_driver_opts):
        _get_oslo_driver_opts.return_value = []
        result = self.driver.get_driver_options()
        actual = (infinidat.infinidat_opts)
        self.assertEqual(actual, result)

    @skip_driver_setup
    def test__setup_and_get_system_object(self):
        # This test should skip the driver setup, as it generates more calls to
        # the add_auto_retry, set_source_identifier and login methods:
        auth = (self.configuration.san_login,
                self.configuration.san_password)

        self.driver._setup_and_get_system_object(
            self.configuration.san_ip, auth)

        self._system.api.add_auto_retry.assert_called_once()
        self._system.api.set_source_identifier.assert_called_once_with(
            infinidat._INFINIDAT_CINDER_IDENTIFIER)
        self._system.login.assert_called_once()

    @skip_driver_setup
    @mock.patch('cinder.volume.drivers.infinidat.infinisdk', None)
    def test_do_setup_no_infinisdk(self):
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.do_setup, None)

    @mock.patch('cinder.volume.drivers.infinidat.infinisdk.InfiniBox')
    @ddt.data(True, False)
    def test_ssl_options(self, use_ssl, infinibox):
        auth = (self.configuration.san_login,
                self.configuration.san_password)
        self.override_config('driver_use_ssl', use_ssl)
        self.driver.do_setup(None)
        infinibox.assert_called_once_with(self.configuration.san_ip,
                                          auth=auth, use_ssl=use_ssl)

    def test_create_export_snapshot(self):
        self.assertIsNone(self.driver.create_export_snapshot(
            None, test_snapshot, test_connector))

    def test_remove_export_snapshot(self):
        self.assertIsNone(self.driver.remove_export_snapshot(
            None, test_snapshot))

    def test_backup_use_temp_snapshot(self):
        self.assertTrue(self.driver.backup_use_temp_snapshot())

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_get_infinidat_snapshot')
    def test_initialize_connection_snapshot(self, get_snapshot):
        result = self.driver.initialize_connection_snapshot(
            test_snapshot, test_connector)
        get_snapshot.assert_called_once_with(test_snapshot)
        self.assertEqual(1, result["data"]["target_lun"])

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_get_infinidat_volume')
    def test_initialize_connection(self, get_volume):
        self._system.hosts.safe_get.return_value = None
        result = self.driver.initialize_connection(test_volume, test_connector)
        get_volume.assert_called_once_with(test_volume)
        self.assertEqual(1, result["data"]["target_lun"])

    def test_initialize_connection_host_exists(self):
        result = self.driver.initialize_connection(test_volume, test_connector)
        self.assertEqual(1, result["data"]["target_lun"])

    def test_initialize_connection_mapping_exists(self):
        mock_mapping = mock.Mock()
        mock_mapping.get_volume.return_value = self._mock_volume
        mock_mapping.get_lun.return_value = 888
        self._mock_host.get_luns.return_value = [mock_mapping]
        result = self.driver.initialize_connection(test_volume, test_connector)
        self.assertEqual(888, result["data"]["target_lun"])

    def test_initialize_connection_mapping_not_found(self):
        mock_mapping = mock.Mock()
        mock_mapping.get_volume.return_value = None
        self._mock_host.get_luns.return_value = [mock_mapping]
        result = self.driver.initialize_connection(test_volume, test_connector)
        self.assertEqual(TEST_LUN, result["data"]["target_lun"])

    def test_initialize_connection_volume_doesnt_exist(self):
        self._system.volumes.safe_get.return_value = None
        self.assertRaises(exception.VolumeNotFound,
                          self.driver.initialize_connection,
                          test_volume, test_connector)

    def test_initialize_connection_create_fails(self):
        self._system.hosts.safe_get.return_value = None
        self._system.hosts.create.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection,
                          test_volume, test_connector)

    def test_initialize_connection_map_fails(self):
        self._mock_host.map_volume.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection,
                          test_volume, test_connector)

    def test_initialize_connection_metadata(self):
        self._system.hosts.safe_get.return_value = None
        self.driver.initialize_connection(test_volume, test_connector)
        self._validate_host_metadata()

    @ddt.data({'connector': None, 'multiattach': True,
               'attachment': [test_attachment1, test_attachment1]},
              {'connector': test_connector3, 'multiattach': True,
               'attachment': [test_attachment1, test_attachment1]},
              {'connector': test_connector, 'multiattach': False,
               'attachment': [test_attachment1]},
              {'connector': test_connector, 'multiattach': True,
               'attachment': None},
              {'connector': test_connector, 'multiattach': True,
               'attachment': [test_attachment2, test_attachment3]})
    @ddt.unpack
    def test__is_volume_multiattached_negative(self, connector,
                                               multiattach, attachment):
        volume = copy.deepcopy(test_volume)
        volume.multiattach = multiattach
        volume.volume_attachment = attachment
        self.assertFalse(self.driver._is_volume_multiattached(volume,
                                                              connector))

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_terminate_connection')
    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_get_infinidat_volume')
    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_is_volume_multiattached')
    def test_terminate_connection(self, volume_multiattached, get_volume,
                                  terminate_connection):
        volume = copy.deepcopy(test_volume)
        volume.volume_attachment = [test_attachment1]
        volume_multiattached.return_value = False
        get_volume.return_value = self._mock_volume
        self.assertFalse(self.driver.terminate_connection(volume,
                                                          test_connector))
        volume_multiattached.assert_called_once_with(volume, test_connector)
        get_volume.assert_called_once_with(volume)
        terminate_connection.assert_called_once_with(self._mock_volume,
                                                     test_connector)

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_terminate_connection')
    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_get_infinidat_snapshot')
    def test_terminate_connection_snapshot(self, get_snapshot,
                                           terminate_connection):
        get_snapshot.return_value = self._mock_snapshot
        self.assertIsNone(self.driver.terminate_connection_snapshot(
            test_snapshot, test_connector))
        get_snapshot.assert_called_once_with(test_snapshot)
        terminate_connection.assert_called_once_with(self._mock_snapshot,
                                                     test_connector)

    def test_terminate_connection_delete_host(self):
        self._mock_host.get_luns.return_value = [object()]
        volume = copy.deepcopy(test_volume)
        volume.volume_attachment = [test_attachment1]
        self.assertFalse(self.driver.terminate_connection(volume,
                                                          test_connector))
        self.assertEqual(0, self._mock_host.safe_delete.call_count)
        self._mock_host.get_luns.return_value = []
        self.assertFalse(self.driver.terminate_connection(volume,
                                                          test_connector))
        self.assertEqual(1, self._mock_host.safe_delete.call_count)

    def test_terminate_connection_volume_doesnt_exist(self):
        self._system.volumes.safe_get.return_value = None
        self.assertRaises(exception.VolumeNotFound,
                          self.driver.terminate_connection,
                          test_volume, test_connector)

    def test_terminate_connection_api_fail(self):
        self._mock_host.unmap_volume.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.terminate_connection,
                          test_volume, test_connector)

    def test_get_volume_stats_refreshes(self):
        result = self.driver.get_volume_stats()
        self.assertEqual(1, result["free_capacity_gb"])
        # change the "free space" in the pool
        self._mock_pool.get_free_physical_capacity.return_value = 0
        # no refresh - free capacity should stay the same
        result = self.driver.get_volume_stats(refresh=False)
        self.assertEqual(1, result["free_capacity_gb"])
        # refresh - free capacity should change to 0
        result = self.driver.get_volume_stats(refresh=True)
        self.assertEqual(0, result["free_capacity_gb"])

    def test_get_volume_stats_pool_not_found(self):
        self._system.pools.safe_get.return_value = None
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.get_volume_stats)

    def test_get_volume_stats_max_over_subscription_ratio(self):
        self.override_config('san_thin_provision', True)
        self.override_config('max_over_subscription_ratio', 10.0)
        result = self.driver.get_volume_stats()
        self.assertEqual('10.0', result['max_over_subscription_ratio'])
        self.assertTrue(result['thin_provisioning_support'])
        self.assertFalse(result['thick_provisioning_support'])

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_volume(self, *mocks):
        self.driver.create_volume(test_volume)

    def test_create_volume_pool_not_found(self):
        self._system.pools.safe_get.return_value = None
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_volume, test_volume)

    def test_create_volume_api_fail(self):
        self._system.pools.safe_get.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume, test_volume)

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_volume_metadata(self, *mocks):
        self.driver.create_volume(test_volume)
        self._validate_object_metadata(self._mock_volume, test_volume)

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_volume_compression_enabled(self, *mocks):
        self.override_config('infinidat_use_compression', True)
        self.driver.create_volume(test_volume)
        self.assertTrue(
            self._system.volumes.create.call_args[1]["compression_enabled"]
        )

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_volume_compression_disabled(self, *mocks):
        self.override_config('infinidat_use_compression', False)
        self.driver.create_volume(test_volume)
        self.assertFalse(
            self._system.volumes.create.call_args[1]["compression_enabled"]
        )

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_volume_compression_default(self, *mocks):
        self.driver.create_volume(test_volume)
        self.assertNotIn(
            "compression_enabled",
            self._system.volumes.create.call_args[1]
        )

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    @mock.patch('cinder.volume.volume_utils.group_get_by_id')
    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_create_volume_within_group(self, *mocks):
        self.driver.create_volume(test_volume3)
        self._mock_group.add_member.assert_called_once()

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    @mock.patch('cinder.volume.volume_utils.group_get_by_id')
    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=False)
    def test_create_volume_within_no_cg_group(self, *mocks):
        self.driver.create_volume(test_volume3)
        self._mock_group.add_member.assert_not_called()

    def test_delete_volume(self):
        self.driver.delete_volume(test_volume)

    def test_delete_volume_doesnt_exist(self):
        self._system.volumes.safe_get.return_value = None
        # should not raise an exception
        self.driver.delete_volume(test_volume)

    def test_delete_volume_with_children(self):
        self._mock_volume.has_children.return_value = True
        self.assertRaises(exception.VolumeIsBusy,
                          self.driver.delete_volume, test_volume)

    def test_extend_volume(self):
        self.driver.extend_volume(test_volume, 2)
        self._mock_volume.resize.assert_called_with(1 * units.Gi)

    def test_extend_volume_api_fail(self):
        self._mock_volume.resize.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.extend_volume, test_volume, 2)

    def test_create_snapshot(self):
        self.driver.create_snapshot(test_snapshot)

    def test_create_snapshot_metadata(self):
        self._mock_volume.create_snapshot.return_value = self._mock_volume
        self.driver.create_snapshot(test_snapshot)
        self._validate_object_metadata(self._mock_volume, test_snapshot)

    def test_create_snapshot_volume_doesnt_exist(self):
        self._system.volumes.safe_get.return_value = None
        self.assertRaises(exception.VolumeNotFound,
                          self.driver.create_snapshot, test_snapshot)

    def test_create_snapshot_api_fail(self):
        self._mock_volume.create_snapshot.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_snapshot, test_snapshot)

    @mock.patch("cinder.volume.volume_utils.copy_volume")
    @mock.patch("cinder.volume.volume_utils.brick_get_connector")
    @mock.patch("cinder.volume.volume_utils.brick_get_connector_properties",
                return_value=test_connector)
    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_volume_from_snapshot(self, *mocks):
        self.driver.create_volume_from_snapshot(test_clone, test_snapshot)

    def test_create_volume_from_snapshot_doesnt_exist(self):
        self._system.volumes.safe_get.return_value = None
        self.assertRaises(exception.SnapshotNotFound,
                          self.driver.create_volume_from_snapshot,
                          test_clone, test_snapshot)

    def test_create_volume_from_snapshot_create_fails(self):
        self._system.volumes.create.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume_from_snapshot,
                          test_clone, test_snapshot)

    @mock.patch("cinder.volume.volume_utils.brick_get_connector_properties",
                return_value=test_connector)
    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_volume_from_snapshot_map_fails(self, *mocks):
        self._mock_host.map_volume.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume_from_snapshot,
                          test_clone, test_snapshot)

    @mock.patch('cinder.volume.volume_utils.brick_get_connector')
    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    @mock.patch('cinder.volume.volume_utils.brick_get_connector_properties')
    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_connect_device')
    def test_create_volume_from_snapshot_connect_fails(self, connect_device,
                                                       connector_properties,
                                                       *mocks):
        connector_properties.return_value = test_connector
        connect_device.side_effect = exception.DeviceUnavailable(
            path='/dev/sdb', reason='Block device required')
        self.assertRaises(exception.DeviceUnavailable,
                          self.driver.create_volume_from_snapshot,
                          test_clone, test_snapshot)

    def test_delete_snapshot(self):
        self.driver.delete_snapshot(test_snapshot)

    def test_delete_snapshot_doesnt_exist(self):
        self._system.volumes.safe_get.return_value = None
        # should not raise an exception
        self.driver.delete_snapshot(test_snapshot)

    def test_delete_snapshot_api_fail(self):
        self._mock_volume.safe_delete.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_snapshot, test_snapshot)

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                'delete_snapshot')
    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                'create_volume_from_snapshot')
    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                'create_snapshot')
    @mock.patch('uuid.uuid4')
    def test_create_cloned_volume(self, mock_uuid, create_snapshot,
                                  create_volume_from_snapshot,
                                  delete_snapshot):
        mock_uuid.return_value = uuid.UUID(test_snapshot.id)
        snapshot_attributes = ('id', 'name', 'volume')
        Snapshot = collections.namedtuple('Snapshot', snapshot_attributes)
        snapshot_id = test_snapshot.id
        snapshot_name = self.configuration.snapshot_name_template % snapshot_id
        snapshot = Snapshot(id=snapshot_id, name=snapshot_name,
                            volume=test_volume)
        self.driver.create_cloned_volume(test_clone, test_volume)
        create_snapshot.assert_called_once_with(snapshot)
        create_volume_from_snapshot.assert_called_once_with(test_clone,
                                                            snapshot)
        delete_snapshot.assert_called_once_with(snapshot)

    def test_create_cloned_volume_create_fails(self):
        self._system.volumes.create.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_cloned_volume,
                          test_clone, test_volume)

    @mock.patch("cinder.volume.volume_utils.brick_get_connector_properties",
                return_value=test_connector)
    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_cloned_volume_map_fails(self, *mocks):
        self._mock_host.map_volume.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_cloned_volume,
                          test_clone, test_volume)

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_create_group(self, *mocks):
        self.driver.create_group(None, test_group)

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=False)
    def test_create_generic_group(self, *mocks):
        self.assertRaises(NotImplementedError,
                          self.driver.create_group,
                          None, test_group)

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_create_group_metadata(self, *mocks):
        self.driver.create_group(None, test_group)
        self._validate_object_metadata(self._mock_group, test_group)

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_create_group_twice(self, *mocks):
        self.driver.create_group(None, test_group)
        self.driver.create_group(None, test_group)

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_create_group_api_fail(self, *mocks):
        self._system.cons_groups.create.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_group,
                          None, test_group)

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_delete_group(self, *mocks):
        self.driver.delete_group(None, test_group, [test_volume])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=False)
    def test_delete_generic_group(self, *mocks):
        self.assertRaises(NotImplementedError,
                          self.driver.delete_group,
                          None, test_group, [test_volume])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_delete_group_doesnt_exist(self, *mocks):
        self._system.cons_groups.safe_get.return_value = None
        self.driver.delete_group(None, test_group, [test_volume])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_delete_group_api_fail(self, *mocks):
        self._mock_group.safe_delete.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_group,
                          None, test_group, [test_volume])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_update_group_add_and_remove(self, *mocks):
        self.driver.update_group(None, test_group,
                                 [test_volume], [test_volume])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=False)
    def test_update_generic_group_add_and_remove(self, *mocks):
        self.assertRaises(NotImplementedError,
                          self.driver.update_group,
                          None, test_group, [test_volume], [test_volume])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_update_group_api_fail(self, *mocks):
        self._mock_group.add_member.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.update_group,
                          None, test_group,
                          [test_volume], [test_volume])

    @mock.patch("cinder.volume.volume_utils.copy_volume")
    @mock.patch("cinder.volume.volume_utils.brick_get_connector")
    @mock.patch("cinder.volume.volume_utils.brick_get_connector_properties",
                return_value=test_connector)
    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_group_from_src_snaps(self, *mocks):
        self.driver.create_group_from_src(None, test_group, [test_volume],
                                          test_snapgroup, [test_snapshot],
                                          None, None)

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=False)
    def test_create_genericgroup_from_src_snaps(self, *mocks):
        self.assertRaises(NotImplementedError,
                          self.driver.create_group_from_src,
                          None, test_group, [test_volume],
                          test_snapgroup, [test_snapshot],
                          None, None)

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_create_group_from_empty_sources(self, *mocks):
        self.assertRaises(exception.InvalidInput,
                          self.driver.create_group_from_src,
                          None, test_group, [test_volume],
                          None, None, None, None)

    @mock.patch("cinder.volume.volume_utils.copy_volume")
    @mock.patch("cinder.volume.volume_utils.brick_get_connector")
    @mock.patch("cinder.volume.volume_utils.brick_get_connector_properties",
                return_value=test_connector)
    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_create_group_from_src_vols(self, *mocks):
        self.driver.create_group_from_src(None, test_group, [test_volume],
                                          None, None,
                                          test_group, [test_volume])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_create_group_snap(self, *mocks):
        mock_snapgroup = mock.Mock()
        mock_snapgroup.get_members.return_value = [self._mock_snapshot,
                                                   self._mock_snapshot]
        self._mock_volume.get_name.side_effect = [fake.VOLUME_NAME,
                                                  fake.VOLUME2_NAME]
        self._mock_group.create_snapshot.return_value = mock_snapgroup
        self.driver.create_group_snapshot(None, test_snapgroup,
                                          [test_snapshot, test_snapshot])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=False)
    def test_create_generic_group_snap(self, *mocks):
        self.assertRaises(NotImplementedError,
                          self.driver.create_group_snapshot,
                          None, test_snapgroup, [test_snapshot])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_create_group_snap_api_fail(self, *mocks):
        self._mock_group.create_snapshot.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_group_snapshot, None,
                          test_snapgroup, [test_snapshot])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_delete_group_snap(self, *mocks):
        self.driver.delete_group_snapshot(None,
                                          test_snapgroup,
                                          [test_snapshot])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=False)
    def test_delete_generic_group_snap(self, *mocks):
        self.assertRaises(NotImplementedError,
                          self.driver.delete_group_snapshot,
                          None, test_snapgroup, [test_snapshot])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_delete_group_snap_does_not_exist(self, *mocks):
        self._system.cons_groups.safe_get.return_value = None
        self.driver.delete_group_snapshot(None,
                                          test_snapgroup,
                                          [test_snapshot])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_delete_group_snap_invalid_group(self, *mocks):
        self._mock_group.is_snapgroup.return_value = False
        self.assertRaises(exception.InvalidGroupSnapshot,
                          self.driver.delete_group_snapshot,
                          None, test_snapgroup, [test_snapshot])

    @mock.patch('cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
                return_value=True)
    def test_delete_group_snap_api_fail(self, *mocks):
        self._mock_group.safe_delete.side_effect = self._raise_infinisdk
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_group_snapshot,
                          None, test_snapgroup, [test_snapshot])

    def test_snapshot_revert_use_temp_snapshot(self):
        result = self.driver.snapshot_revert_use_temp_snapshot()
        self.assertFalse(result)

    @ddt.data((1, 1), (1, 2))
    @ddt.unpack
    def test_revert_to_snapshot_resize(self, volume_size, snapshot_size):
        volume = copy.deepcopy(test_volume)
        snapshot = copy.deepcopy(test_snapshot)
        snapshot.volume.size = snapshot_size
        self._system.volumes.safe_get.side_effect = [self._mock_snapshot,
                                                     self._mock_volume,
                                                     self._mock_volume]
        self._mock_volume.get_size.side_effect = [volume_size * units.Gi,
                                                  volume_size * units.Gi]
        self.driver.revert_to_snapshot(None, volume, snapshot)
        self._mock_volume.restore.assert_called_once_with(self._mock_snapshot)
        if volume_size == snapshot_size:
            self._mock_volume.resize.assert_not_called()
        else:
            delta = (snapshot_size - volume_size) * units.Gi
            self._mock_volume.resize.assert_called_with(delta)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_by_source_name(self, *mocks):
        existing_ref = {'source-name': TEST_VOLUME_SOURCE_NAME}
        self.driver.manage_existing(test_volume, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_by_source_id(self, *mocks):
        existing_ref = {'source-id': TEST_VOLUME_SOURCE_ID}
        self.driver.manage_existing(test_volume, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_by_invalid_source(self, *mocks):
        existing_ref = {'source-path': None}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing,
                          test_volume, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    @mock.patch('cinder.volume.volume_utils.check_already_managed_volume',
                return_value=False)
    def test_manage_existing_not_managed(self, *mocks):
        self._mock_volume.get_all_metadata.return_value = (
            TEST_VOLUME_METADATA)
        existing_ref = {'source-name': TEST_VOLUME_SOURCE_NAME}
        self.driver.manage_existing(test_volume, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    @mock.patch('cinder.volume.volume_utils.check_already_managed_volume',
                return_value=True)
    def test_manage_existing_already_managed(self, *mocks):
        self._mock_volume.get_all_metadata.return_value = (
            TEST_VOLUME_METADATA)
        existing_ref = {'source-name': TEST_VOLUME_SOURCE_NAME}
        self.assertRaises(exception.ManageExistingAlreadyManaged,
                          self.driver.manage_existing,
                          test_volume, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_invalid_pool(self, *mocks):
        existing_ref = {'source-name': TEST_VOLUME_SOURCE_NAME}
        self._mock_volume.get_pool_name.return_value = TEST_POOL_NAME2
        self.assertRaises(exception.InvalidConfigurationValue,
                          self.driver.manage_existing,
                          test_volume, existing_ref)

    def test_manage_existing_get_size(self):
        existing_ref = {'source-name': TEST_VOLUME_SOURCE_NAME}
        size = self.driver.manage_existing_get_size(test_volume, existing_ref)
        self.assertEqual(test_volume.size, size)

    def test_get_manageable_volumes(self):
        cinder_volumes = [test_volume]
        self._mock_volume.is_snapshot.return_value = False
        self._mock_volume.get_all_metadata.return_value = {
            'cinder_id': fake.VOLUME2_ID
        }
        self.driver.get_manageable_volumes(cinder_volumes, None,
                                           1, 0, [], [])

    def test_get_manageable_volumes_already_managed(self):
        cinder_volumes = [test_volume]
        self._mock_volume.get_id.return_value = TEST_VOLUME_SOURCE_ID
        self._mock_volume.get_all_metadata.return_value = (
            TEST_VOLUME_METADATA)
        self._mock_volume.is_snapshot.return_value = False
        self.driver.get_manageable_volumes(cinder_volumes, None,
                                           1, 0, [], [])

    def test_get_manageable_volumes_but_snapshots(self):
        cinder_volumes = [test_volume]
        self._mock_volume.is_snapshot.return_value = True
        self.driver.get_manageable_volumes(cinder_volumes, None,
                                           1, 0, [], [])

    def test_get_manageable_volumes_has_mappings(self):
        cinder_volumes = [test_volume]
        self._mock_volume.is_snapshot.return_value = False
        self._mock_volume.get_all_metadata.return_value = {
            'cinder_id': fake.VOLUME2_ID
        }
        lun = mock.Mock()
        self._mock_volume.get_logical_units.return_value = [lun]
        self.driver.get_manageable_volumes(cinder_volumes, None,
                                           1, 0, [], [])

    def test_get_manageable_volumes_has_snapshots(self):
        cinder_volumes = [test_volume]
        self._mock_volume.is_snapshot.return_value = False
        self._mock_volume.has_children.return_value = True
        self._mock_volume.get_all_metadata.return_value = {
            'cinder_id': fake.VOLUME2_ID
        }
        self.driver.get_manageable_volumes(cinder_volumes, None,
                                           1, 0, [], [])

    def test_unmanage(self):
        self.driver.unmanage(test_volume)

    @mock.patch('cinder.objects.Snapshot.exists', return_value=True)
    def test__check_already_managed_snapshot(self, *mocks):
        self.driver._check_already_managed_snapshot(test_snapshot.id)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_snapshot_by_source_name(self, *mocks):
        existing_ref = {'source-name': TEST_SNAPSHOT_SOURCE_NAME}
        self.driver.manage_existing_snapshot(test_snapshot, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_snapshot_by_source_id(self, *mocks):
        existing_ref = {'source-id': TEST_SNAPSHOT_SOURCE_ID}
        self.driver.manage_existing_snapshot(test_snapshot, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_snapshot_but_volume(self, *mocks):
        existing_ref = {'source-id': TEST_SNAPSHOT_SOURCE_ID}
        self._mock_volume.is_snapshot.return_value = False
        self.assertRaises(exception.InvalidSnapshot,
                          self.driver.manage_existing_snapshot,
                          test_snapshot, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_snapshot_by_invalid_source(self, *mocks):
        existing_ref = {'source-path': None}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot,
                          test_snapshot, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_snapshot_by_non_cinder_id(self, *mocks):
        self._mock_volume.get_all_metadata.return_value = {'cinder_id': 'x'}
        existing_ref = {'source-id': TEST_SNAPSHOT_SOURCE_ID}
        self.driver.manage_existing_snapshot(test_snapshot, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_check_already_managed_snapshot', return_value=False)
    def test_manage_existing_snapshot_not_managed(self, *mocks):
        self._mock_volume.get_all_metadata.return_value = (
            TEST_SNAPSHOT_METADATA)
        existing_ref = {'source-name': TEST_SNAPSHOT_SOURCE_NAME}
        self.driver.manage_existing(test_snapshot, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_check_already_managed_snapshot', return_value=True)
    def test_manage_existing_snapshot_already_managed(self, *mocks):
        self._mock_volume.get_all_metadata.return_value = (
            TEST_SNAPSHOT_METADATA)
        existing_ref = {'source-name': TEST_SNAPSHOT_SOURCE_NAME}
        self.assertRaises(exception.ManageExistingAlreadyManaged,
                          self.driver.manage_existing_snapshot,
                          test_snapshot, existing_ref)

    @mock.patch('cinder.volume.volume_types.get_volume_type_qos_specs')
    def test_manage_existing_snapshot_invalid_pool(self, *mocks):
        existing_ref = {'source-name': TEST_SNAPSHOT_SOURCE_NAME}
        self._mock_volume.get_pool_name.return_value = TEST_POOL_NAME2
        self.assertRaises(exception.InvalidConfigurationValue,
                          self.driver.manage_existing_snapshot,
                          test_snapshot, existing_ref)

    def test_manage_existing_snapshot_get_size(self):
        existing_ref = {'source-name': TEST_SNAPSHOT_SOURCE_NAME}
        size = self.driver.manage_existing_snapshot_get_size(test_volume,
                                                             existing_ref)
        self.assertEqual(test_snapshot.volume.size, size)

    def test_get_manageable_snapshots(self):
        cinder_snapshots = [test_snapshot]
        self._mock_volume.is_snapshot.return_value = True
        self._mock_volume.get_all_metadata.return_value = {
            'cinder_id': fake.SNAPSHOT2_ID
        }
        self.driver.get_manageable_snapshots(cinder_snapshots,
                                             None, 1, 0, [], [])

    def test_get_manageable_snapshots_already_managed(self):
        cinder_snapshots = [test_snapshot]
        self._mock_volume.get_id.return_value = TEST_SNAPSHOT_SOURCE_ID
        self._mock_volume.get_all_metadata.return_value = (
            TEST_SNAPSHOT_METADATA)
        self._mock_volume.is_snapshot.return_value = True
        self.driver.get_manageable_snapshots(cinder_snapshots,
                                             None, 1, 0, [], [])

    def test_get_manageable_snapshots_but_volumes(self):
        cinder_snapshots = [test_snapshot]
        self._mock_volume.is_snapshot.return_value = False
        self.driver.get_manageable_snapshots(cinder_snapshots,
                                             None, 1, 0, [], [])

    def test_get_manageable_snapshots_has_mappings(self):
        cinder_snapshots = [test_snapshot]
        self._mock_volume.is_snapshot.return_value = True
        self._mock_volume.get_all_metadata.return_value = {
            'cinder_id': fake.SNAPSHOT2_ID
        }
        lun = mock.Mock()
        self._mock_volume.get_logical_units.return_value = [lun]
        self.driver.get_manageable_snapshots(cinder_snapshots,
                                             None, 1, 0, [], [])

    def test_get_manageable_snapshots_has_clones(self):
        cinder_snapshots = [test_snapshot]
        self._mock_volume.is_snapshot.return_value = True
        self._mock_volume.has_children.return_value = True
        self._mock_volume.get_all_metadata.return_value = {
            'cinder_id': fake.SNAPSHOT2_ID
        }
        self.driver.get_manageable_snapshots(cinder_snapshots,
                                             None, 1, 0, [], [])

    def test_unmanage_snapshot(self):
        self.driver.unmanage_snapshot(test_snapshot)

    def test_terminate_connection_no_attachment_connector(self):
        volume = copy.deepcopy(test_volume)
        volume.multiattach = True
        volume.volume_attachment = [test_attachment3]
        self.assertFalse(self.driver.terminate_connection(volume,
                                                          test_connector))

    def test_terminate_connection_no_host(self):
        self._system.hosts.safe_get.return_value = None
        volume = copy.deepcopy(test_volume)
        volume.volume_attachment = [test_attachment1]
        self.assertFalse(self.driver.terminate_connection(volume,
                                                          test_connector))

    def test_terminate_connection_no_mapping(self):
        self._mock_host.unmap_volume.side_effect = KeyError
        volume = copy.deepcopy(test_volume)
        volume.volume_attachment = [test_attachment1]
        self.assertFalse(self.driver.terminate_connection(volume,
                                                          test_connector))

    def test_update_migrated_volume_new_volume_not_found(self):
        self._system.volumes.safe_get.side_effect = [
            None, self._mock_volume]
        self.assertRaises(exception.VolumeNotFound,
                          self.driver.update_migrated_volume,
                          None, test_volume, test_volume2,
                          'available')

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_set_cinder_object_metadata')
    def test_update_migrated_volume_volume_not_found(self, set_metadata):
        self._system.volumes.safe_get.side_effect = [
            self._mock_new_volume, None]
        update = self.driver.update_migrated_volume(None,
                                                    test_volume,
                                                    test_volume2,
                                                    'available')
        expected = {'_name_id': None, 'provider_location': None}
        self.assertEqual(expected, update)
        set_metadata.assert_called_once_with(self._mock_new_volume,
                                             test_volume)

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_set_cinder_object_metadata')
    def test_update_migrated_new_volume_rename_error(self, set_metadata):
        self._system.volumes.safe_get.side_effect = [
            self._mock_new_volume, None]
        self._mock_new_volume.update_name.side_effect = [
            FakeInfinisdkException]
        update = self.driver.update_migrated_volume(None,
                                                    test_volume,
                                                    test_volume2,
                                                    'available')
        expected = {'_name_id': test_volume2.name_id,
                    'provider_location': None}
        self.assertEqual(expected, update)
        set_metadata.assert_called_once_with(self._mock_new_volume,
                                             test_volume)

    @mock.patch('cinder.volume.drivers.infinidat.InfiniboxVolumeDriver.'
                '_set_cinder_object_metadata')
    def test_update_migrated(self, set_metadata):
        self._system.volumes.safe_get.side_effect = [
            self._mock_new_volume, self._mock_volume]
        self._mock_new_volume.update_name.side_effect = None
        update = self.driver.update_migrated_volume(None,
                                                    test_volume,
                                                    test_volume2,
                                                    'available')
        expected = {'_name_id': test_volume2.name_id,
                    'provider_location': None}
        self.assertEqual(expected, update)
        set_metadata.assert_called_once_with(self._mock_new_volume,
                                             test_volume)
        self.assertEqual(0, self._log.error.call_count)

    @ddt.data(None, {})
    def test_migrate_volume_no_host(self, host):
        expected = False, None
        update = self.driver.migrate_volume(None, test_volume, host)
        self.assertEqual(expected, update)

    @ddt.data(None, {})
    def test_migrate_volume_no_capabilities(self, capabilities):
        expected = False, None
        host = {'capabilities': capabilities}
        update = self.driver.migrate_volume(None, test_volume, host)
        self.assertEqual(expected, update)

    @ddt.data(None, 123, 'location')
    def test_migrate_volume_invalid_location_info(self, location_info):
        expected = False, None
        capabilities = {'location_info': location_info}
        host = {'capabilities': capabilities}
        update = self.driver.migrate_volume(None, test_volume, host)
        self.assertEqual(expected, update)

    def test_migrate_volume_invalid_driver(self):
        expected = False, None
        location_info = 'vendor:0:/path'
        capabilities = {'location_info': location_info}
        host = {'capabilities': capabilities}
        update = self.driver.migrate_volume(None, test_volume, host)
        self.assertEqual(expected, update)

    def test_migrate_volume_invalid_serial(self):
        expected = False, None
        location_info = '%s:%s:%s' % (self.driver.__class__.__name__,
                                      TEST_SYSTEM_SERIAL2, TEST_POOL_NAME2)
        capabilities = {'location_info': location_info}
        host = {'capabilities': capabilities}
        update = self.driver.migrate_volume(None, test_volume, host)
        self.assertEqual(expected, update)

    def test_migrate_volume_same_pool(self):
        expected = True, None
        location_info = '%s:%s:%s' % (self.driver.__class__.__name__,
                                      TEST_SYSTEM_SERIAL, TEST_POOL_NAME)
        capabilities = {'location_info': location_info}
        host = {'capabilities': capabilities}
        update = self.driver.migrate_volume(None, test_volume, host)
        self.assertEqual(expected, update)

    def test_migrate_volume_no_pool(self):
        expected = False, None
        self._system.pools.safe_get.return_value = None
        location_info = '%s:%s:%s' % (self.driver.__class__.__name__,
                                      TEST_SYSTEM_SERIAL, TEST_POOL_NAME2)
        capabilities = {'location_info': location_info}
        host = {'capabilities': capabilities}
        update = self.driver.migrate_volume(None, test_volume, host)
        self.assertEqual(expected, update)

    def test_migrate_volume(self):
        expected = True, None
        location_info = '%s:%s:%s' % (self.driver.__class__.__name__,
                                      TEST_SYSTEM_SERIAL, TEST_POOL_NAME2)
        capabilities = {'location_info': location_info}
        host = {'capabilities': capabilities}
        update = self.driver.migrate_volume(None, test_volume, host)
        self.assertEqual(expected, update)


@ddt.ddt
class InfiniboxDriverTestCaseFC(InfiniboxDriverTestCaseBase):
    @ddt.data(*itertools.product(('UP', 'DOWN'), ('OK', 'ERROR')))
    @ddt.unpack
    def test_initialize_connection_nodes_ports(self, link_state, port_state):
        node = mock.Mock()
        port = mock.Mock()
        port.get_link_state.return_value = link_state
        port.get_state.return_value = port_state
        node.get_fc_ports.return_value = [port]
        self._system.components.nodes.get_all.return_value = [node]
        result = self.driver.initialize_connection(test_volume, test_connector)
        self.assertEqual(1, result["data"]["target_lun"])

    def test_initialize_connection_multiple_wwpns(self):
        connector = {'wwpns': [TEST_WWN_1, TEST_WWN_2]}
        result = self.driver.initialize_connection(test_volume, connector)
        self.assertEqual(1, result["data"]["target_lun"])

    def test_validate_connector(self):
        fc_connector = {'wwpns': [TEST_WWN_1, TEST_WWN_2]}
        iscsi_connector = {'initiator': TEST_INITIATOR_IQN}
        self.driver.validate_connector(fc_connector)
        self.assertRaises(exception.InvalidConnectorException,
                          self.driver.validate_connector, iscsi_connector)

    @ddt.data({'connector': test_connector,
               'attachment': [test_attachment1, test_attachment1]},
              {'connector': test_connector2,
               'attachment': [test_attachment2, test_attachment2]})
    @ddt.unpack
    def test__is_volume_multiattached_positive(self, connector, attachment):
        volume = copy.deepcopy(test_volume)
        volume.multiattach = True
        volume.volume_attachment = attachment
        self.assertTrue(self.driver._is_volume_multiattached(volume,
                                                             connector))

    def test_terminate_connection_multiattached_volume(self):
        volume = copy.deepcopy(test_volume)
        volume.multiattach = True
        volume.volume_attachment = [test_attachment1, test_attachment1]
        self.assertTrue(self.driver.terminate_connection(volume,
                                                         test_connector))

    def test_terminate_connection_force_detach(self):
        mock_infinidat_host = mock.Mock()
        mock_infinidat_host.get_ports.return_value = [
            self._wwn.WWN(TEST_WWN_1)]
        mock_mapping = mock.Mock()
        mock_mapping.get_host.return_value = mock_infinidat_host
        self._mock_volume.get_logical_units.return_value = [mock_mapping]
        volume = copy.deepcopy(test_volume)
        volume.volume_attachment = [test_attachment1, test_attachment2]
        self.assertTrue(self.driver.terminate_connection(volume, None))
        self._mock_host.unmap_volume.assert_called_once()
        self._mock_host.safe_delete.assert_called_once()


@ddt.ddt
class InfiniboxDriverTestCaseISCSI(InfiniboxDriverTestCaseBase):
    def setUp(self):
        super(InfiniboxDriverTestCaseISCSI, self).setUp()
        self.override_config('infinidat_storage_protocol',
                             TEST_ISCSI_PROTOCOL)
        self.override_config('infinidat_iscsi_netspaces',
                             [TEST_ISCSI_NAMESPACE1])
        self.override_config('use_chap_auth', False)
        self.driver.do_setup(None)

    def test_setup_without_netspaces_configured(self):
        self.override_config('infinidat_iscsi_netspaces', [])
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.do_setup, None)

    def test_initialize_connection(self):
        result = self.driver.initialize_connection(test_volume, test_connector)
        expected = {
            'driver_volume_type': TEST_ISCSI_PROTOCOL,
            'data': {
                'target_discovered': True,
                'target_portal': TEST_TARGET_PORTAL1,
                'target_iqn': TEST_TARGET_IQN,
                'target_lun': TEST_LUN,
                'target_portals': [
                    TEST_TARGET_PORTAL1
                ],
                'target_iqns': [
                    TEST_TARGET_IQN
                ],
                'target_luns': [
                    TEST_LUN
                ]
            }
        }
        self.assertEqual(expected, result)

    def test_initialize_netspace_does_not_exist(self):
        self._system.network_spaces.safe_get.return_value = None
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.initialize_connection,
                          test_volume, test_connector)

    def test_initialize_netspace_has_no_ips(self):
        self._mock_name_space1.get_ips.return_value = []
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.initialize_connection,
                          test_volume, test_connector)

    def test_initialize_connection_with_chap(self):
        self.override_config('use_chap_auth', True)
        result = self.driver.initialize_connection(test_volume, test_connector)
        self.assertEqual(1, result['data']['target_lun'])
        self.assertEqual('CHAP', result['data']['auth_method'])
        self.assertIn('auth_username', result['data'])
        self.assertIn('auth_password', result['data'])

    def test_initialize_connection_multiple_netspaces(self):
        self.override_config('infinidat_iscsi_netspaces',
                             [TEST_ISCSI_NAMESPACE1, TEST_ISCSI_NAMESPACE2])
        self._system.network_spaces.safe_get.side_effect = [
            self._mock_name_space1, self._mock_name_space2]
        result = self.driver.initialize_connection(test_volume, test_connector)
        expected = {
            'driver_volume_type': TEST_ISCSI_PROTOCOL,
            'data': {
                'target_discovered': True,
                'target_portal': TEST_TARGET_PORTAL1,
                'target_iqn': TEST_TARGET_IQN,
                'target_lun': TEST_LUN,
                'target_portals': [
                    TEST_TARGET_PORTAL1,
                    TEST_TARGET_PORTAL3
                ],
                'target_iqns': [
                    TEST_TARGET_IQN,
                    TEST_TARGET_IQN
                ],
                'target_luns': [
                    TEST_LUN,
                    TEST_LUN
                ]
            }
        }
        self.assertEqual(expected, result)

    def test_initialize_connection_multiple_netspaces_multipath(self):
        self.override_config('infinidat_iscsi_netspaces',
                             [TEST_ISCSI_NAMESPACE1, TEST_ISCSI_NAMESPACE2])
        self._system.network_spaces.safe_get.side_effect = [
            self._mock_name_space1, self._mock_name_space2]
        self._mock_name_space1.get_ips.return_value = [
            mock.Mock(ip_address=TEST_IP_ADDRESS1, enabled=True),
            mock.Mock(ip_address=TEST_IP_ADDRESS2, enabled=True)]
        self._mock_name_space2.get_ips.return_value = [
            mock.Mock(ip_address=TEST_IP_ADDRESS3, enabled=True),
            mock.Mock(ip_address=TEST_IP_ADDRESS4, enabled=True)]
        result = self.driver.initialize_connection(test_volume, test_connector)
        expected = {
            'driver_volume_type': TEST_ISCSI_PROTOCOL,
            'data': {
                'target_discovered': True,
                'target_portal': TEST_TARGET_PORTAL1,
                'target_iqn': TEST_TARGET_IQN,
                'target_lun': TEST_LUN,
                'target_portals': [
                    TEST_TARGET_PORTAL1,
                    TEST_TARGET_PORTAL2,
                    TEST_TARGET_PORTAL3,
                    TEST_TARGET_PORTAL4
                ],
                'target_iqns': [
                    TEST_TARGET_IQN,
                    TEST_TARGET_IQN,
                    TEST_TARGET_IQN,
                    TEST_TARGET_IQN
                ],
                'target_luns': [
                    TEST_LUN,
                    TEST_LUN,
                    TEST_LUN,
                    TEST_LUN
                ]
            }
        }
        self.assertEqual(expected, result)

    def test_initialize_connection_disabled_interface(self):
        self._mock_name_space1.get_ips.return_value = [
            mock.Mock(ip_address=TEST_IP_ADDRESS1, enabled=False),
            mock.Mock(ip_address=TEST_IP_ADDRESS2, enabled=True)]
        result = self.driver.initialize_connection(test_volume, test_connector)
        expected = {
            'driver_volume_type': TEST_ISCSI_PROTOCOL,
            'data': {
                'target_discovered': True,
                'target_portal': TEST_TARGET_PORTAL2,
                'target_iqn': TEST_TARGET_IQN,
                'target_lun': TEST_LUN,
                'target_portals': [
                    TEST_TARGET_PORTAL2
                ],
                'target_iqns': [
                    TEST_TARGET_IQN
                ],
                'target_luns': [
                    TEST_LUN
                ]
            }
        }
        self.assertEqual(expected, result)

    def test_initialize_connection_multiple_interfaces(self):
        self._mock_name_space1.get_ips.return_value = [
            mock.Mock(ip_address=TEST_IP_ADDRESS1, enabled=True),
            mock.Mock(ip_address=TEST_IP_ADDRESS2, enabled=True)]
        self._mock_name_space1.get_properties.return_value = mock.Mock(
            iscsi_iqn=TEST_TARGET_IQN, iscsi_tcp_port=TEST_ISCSI_TCP_PORT1)
        result = self.driver.initialize_connection(test_volume, test_connector)
        expected = {
            'driver_volume_type': TEST_ISCSI_PROTOCOL,
            'data': {
                'target_discovered': True,
                'target_portal': TEST_TARGET_PORTAL1,
                'target_iqn': TEST_TARGET_IQN,
                'target_lun': TEST_LUN,
                'target_portals': [
                    TEST_TARGET_PORTAL1,
                    TEST_TARGET_PORTAL2
                ],
                'target_iqns': [
                    TEST_TARGET_IQN,
                    TEST_TARGET_IQN
                ],
                'target_luns': [
                    TEST_LUN,
                    TEST_LUN
                ]
            }
        }
        self.assertEqual(expected, result)

    @ddt.data({'connector': test_connector,
               'attachment': [test_attachment1, test_attachment1]},
              {'connector': test_connector2,
               'attachment': [test_attachment2, test_attachment2]})
    @ddt.unpack
    def test__is_volume_multiattached_positive(self, connector, attachment):
        volume = copy.deepcopy(test_volume)
        volume.multiattach = True
        volume.volume_attachment = attachment
        self.assertTrue(self.driver._is_volume_multiattached(volume,
                                                             connector))

    def test_terminate_connection(self):
        volume = copy.deepcopy(test_volume)
        volume.volume_attachment = [test_attachment1]
        self.assertFalse(self.driver.terminate_connection(volume,
                                                          test_connector))

    def test_terminate_connection_force_detach(self):
        mock_infinidat_host = mock.Mock()
        mock_infinidat_host.get_ports.return_value = [
            self._iqn.IQN(TEST_TARGET_IQN)]
        mock_mapping = mock.Mock()
        mock_mapping.get_host.return_value = mock_infinidat_host
        self._mock_volume.get_logical_units.return_value = [mock_mapping]
        volume = copy.deepcopy(test_volume)
        volume.volume_attachment = [test_attachment1, test_attachment2]
        self.assertTrue(self.driver.terminate_connection(volume, None))
        self._mock_host.unmap_volume.assert_called_once()
        self._mock_host.safe_delete.assert_called_once()

    def test_validate_connector(self):
        fc_connector = {'wwpns': [TEST_WWN_1, TEST_WWN_2]}
        iscsi_connector = {'initiator': TEST_INITIATOR_IQN}
        self.driver.validate_connector(iscsi_connector)
        self.assertRaises(exception.InvalidConnectorException,
                          self.driver.validate_connector, fc_connector)


class InfiniboxDriverTestCaseQoS(InfiniboxDriverTestCaseBase):
    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_no_qos(self, qos_specs):
        qos_specs.return_value = None
        self.driver.create_volume(test_volume)
        self._system.qos_policies.create.assert_not_called()
        self._mock_qos_policy.assign_entity.assert_not_called()

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_qos_max_ipos(self, qos_specs):
        qos_specs.return_value = {'qos_specs': {'id': 'qos_name',
                                                'consumer': 'back-end',
                                                'specs': {'maxIOPS': 1000,
                                                          'maxBWS': None}}}
        self.driver.create_volume(test_volume)
        self._system.qos_policies.create.assert_called_once()
        self._mock_qos_policy.assign_entity.assert_called_once()

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_qos_max_bws(self, qos_specs):
        qos_specs.return_value = {'qos_specs': {'id': 'qos_name',
                                                'consumer': 'back-end',
                                                'specs': {'maxIOPS': None,
                                                          'maxBWS': 10000}}}
        self.driver.create_volume(test_volume)
        self._system.qos_policies.create.assert_called_once()
        self._mock_qos_policy.assign_entity.assert_called_once()

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_qos_no_compat(self, qos_specs):
        qos_specs.return_value = {'qos_specs': {'id': 'qos_name',
                                                'consumer': 'back-end',
                                                'specs': {'maxIOPS': 1000,
                                                          'maxBWS': 10000}}}
        self._system.compat.has_qos.return_value = False
        self.driver.create_volume(test_volume)
        self._system.qos_policies.create.assert_not_called()
        self._mock_qos_policy.assign_entity.assert_not_called()

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_qos_volume_type_id_none(self, qos_specs):
        qos_specs.return_value = {'qos_specs': {'id': 'qos_name',
                                                'consumer': 'back-end',
                                                'specs': {'maxIOPS': 1000,
                                                          'maxBWS': 10000}}}
        self.driver.create_volume(test_volume2)
        self._system.qos_policies.create.assert_not_called()
        self._mock_qos_policy.assign_entity.assert_not_called()

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_qos_no_specs(self, qos_specs):
        qos_specs.return_value = {'qos_specs': None}
        self.driver.create_volume(test_volume)
        self._system.qos_policies.create.assert_not_called()
        self._mock_qos_policy.assign_entity.assert_not_called()

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_qos_front_end(self, qos_specs):
        qos_specs.return_value = {'qos_specs': {'id': 'qos_name',
                                                'consumer': 'front-end',
                                                'specs': {'maxIOPS': 1000,
                                                          'maxBWS': 10000}}}
        self.driver.create_volume(test_volume)
        self._system.qos_policies.create.assert_not_called()
        self._mock_qos_policy.assign_entity.assert_not_called()

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_qos_specs_empty(self, qos_specs):
        qos_specs.return_value = {'qos_specs': {'id': 'qos_name',
                                                'consumer': 'back-end',
                                                'specs': {'maxIOPS': None,
                                                          'maxBWS': None}}}
        self.driver.create_volume(test_volume)
        self._system.qos_policies.create.assert_not_called()
        self._mock_qos_policy.assign_entity.assert_not_called()

    @mock.patch("cinder.volume.volume_types.get_volume_type_qos_specs")
    def test_qos_policy_exists(self, qos_specs):
        qos_specs.return_value = {'qos_specs': {'id': 'qos_name',
                                                'consumer': 'back-end',
                                                'specs': {'maxIOPS': 1000,
                                                          'maxBWS': 10000}}}
        self._system.qos_policies.safe_get.return_value = self._mock_qos_policy
        self.driver.create_volume(test_volume)
        self._system.qos_policies.create.assert_not_called()
        self._mock_qos_policy.assign_entity.assert_called()



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions for all Cinder code.

   This file is for utilities useful in all of Cinder,
   including cinder-manage, the api service, the scheduler,
   etc.

   Code related to volume drivers and connecting to volumes
   should be placed in volume_utils instead.
"""

from collections import OrderedDict
import contextlib
import datetime
import functools
import inspect
import logging as py_logging
import math
import multiprocessing
import operator
import os
import pyclbr
import re
import shutil
import stat
import sys
import tempfile
import typing
from typing import Callable, Iterable, Iterator
from typing import Optional, Type, Union

import eventlet
from eventlet import tpool
from oslo_concurrency import lockutils
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_utils import strutils
from oslo_utils import timeutils
import tenacity

from cinder import coordination
from cinder import exception
from cinder.i18n import _

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
ISO_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
PERFECT_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
INITIAL_AUTO_MOSR = 20
INFINITE_UNKNOWN_VALUES = ('infinite', 'unknown')


synchronized = lockutils.synchronized_with_prefix('cinder-')
synchronized_remove = lockutils.remove_external_lock_file_with_prefix(
    'cinder-')


def clean_volume_file_locks(volume_id, driver):
    """Remove file locks used by Cinder.

    This doesn't take care of driver locks, those should be handled in driver's
    delete_volume method.
    """
    for name in (volume_id + '-delete_volume', volume_id,
                 volume_id + '-detach_volume'):
        try:
            synchronized_remove(name)
        except Exception as exc:
            LOG.warning('Failed to cleanup volume lock %(name)s: %(exc)s',
                        {'name': name, 'exc': exc})

    try:
        driver.clean_volume_file_locks(volume_id)
    except Exception as exc:
        LOG.warning('Failed to cleanup driver locks for volume %(id)s: '
                    '%(exc)s', {'id': volume_id, 'exc': exc})


def api_clean_volume_file_locks(volume_id):
    coordination.synchronized_remove('attachment_update-' + volume_id + '-*')


def clean_snapshot_file_locks(snapshot_id, driver):
    try:
        name = snapshot_id + '-delete_snapshot'
        synchronized_remove(name)
    except Exception as exc:
        LOG.warning('Failed to cleanup snapshot lock %(name)s: %(exc)s',
                    {'name': name, 'exc': exc})

    try:
        driver.clean_snapshot_file_locks(snapshot_id)
    except Exception as exc:
        LOG.warning('Failed to cleanup driver locks for snapshot %(id)s: '
                    '%(exc)s', {'id': snapshot_id, 'exc': exc})


def as_int(obj: Union[int, float, str], quiet: bool = True) -> int:
    # Try "2" -> 2
    try:
        return int(obj)
    except (ValueError, TypeError):
        pass
    # Try "2.5" -> 2
    try:
        return int(float(obj))
    except (ValueError, TypeError):
        pass
    # Eck, not sure what this is then.
    if not quiet:
        raise TypeError(_("Can not translate %s to integer.") % (obj))

    obj = typing.cast(int, obj)
    return obj


def check_exclusive_options(
        **kwargs: Optional[Union[dict, str, bool]]) -> None:
    """Checks that only one of the provided options is actually not-none.

    Iterates over all the kwargs passed in and checks that only one of said
    arguments is not-none, if more than one is not-none then an exception will
    be raised with the names of those arguments who were not-none.
    """

    if not kwargs:
        return

    pretty_keys = kwargs.pop("pretty_keys", True)
    exclusive_options = {}
    for (k, v) in kwargs.items():
        if v is not None:
            exclusive_options[k] = True

    if len(exclusive_options) > 1:
        # Change the format of the names from pythonic to
        # something that is more readable.
        #
        # Ex: 'the_key' -> 'the key'
        if pretty_keys:
            tnames = [k.replace('_', ' ') for k in kwargs]
        else:
            tnames = list(kwargs.keys())
        names = ", ".join(sorted(tnames))
        msg = (_("May specify only one of %s") % (names))
        raise exception.InvalidInput(reason=msg)


def execute(*cmd: str, **kwargs: Union[bool, str]) -> tuple[str, str]:
    """Convenience wrapper around oslo's execute() method."""
    if 'run_as_root' in kwargs and 'root_helper' not in kwargs:
        kwargs['root_helper'] = get_root_helper()
    return processutils.execute(*cmd, **kwargs)


def check_ssh_injection(cmd_list: list[str]) -> None:
    ssh_injection_pattern: tuple[str, ...] = ('`', '$', '|', '||', ';', '&',
                                              '&&', '>', '>>', '<')

    # Check whether injection attacks exist
    for arg in cmd_list:
        arg = arg.strip()

        # Check for matching quotes on the ends
        is_quoted = re.match('^(?P<quote>[\'"])(?P<quoted>.*)(?P=quote)$', arg)
        if is_quoted:
            # Check for unescaped quotes within the quoted argument
            quoted = is_quoted.group('quoted')
            if quoted:
                if (re.match('[\'"]', quoted) or
                        re.search('[^\\\\][\'"]', quoted)):
                    raise exception.SSHInjectionThreat(command=cmd_list)
        else:
            # We only allow spaces within quoted arguments, and that
            # is the only special character allowed within quotes
            if len(arg.split()) > 1:
                raise exception.SSHInjectionThreat(command=cmd_list)

        # Second, check whether danger character in command. So the shell
        # special operator must be a single argument.
        for c in ssh_injection_pattern:
            if c not in arg:
                continue

            result = arg.find(c)
            if not result == -1:
                if result == 0 or not arg[result - 1] == '\\':
                    raise exception.SSHInjectionThreat(command=cmd_list)


def check_metadata_properties(
        metadata: Optional[dict[str, str]]) -> None:
    """Checks that the volume metadata properties are valid."""

    if not metadata:
        metadata = {}
    if not isinstance(metadata, dict):
        msg = _("Metadata should be a dict.")
        raise exception.InvalidInput(msg)

    for k, v in metadata.items():
        try:
            check_string_length(k, "Metadata key: %s" % k, min_length=1)
            check_string_length(v, "Value for metadata key: %s" % k)
        except exception.InvalidInput as exc:
            raise exception.InvalidVolumeMetadata(reason=exc)
        # for backward compatibility
        if len(k) > 255:
            msg = _("Metadata property key %s greater than 255 "
                    "characters.") % k
            raise exception.InvalidVolumeMetadataSize(reason=msg)
        if len(v) > 255:
            msg = _("Metadata property key %s value greater than "
                    "255 characters.") % k
            raise exception.InvalidVolumeMetadataSize(reason=msg)


def last_completed_audit_period(unit: Optional[str] = None) -> \
        tuple[Union[datetime.datetime, datetime.timedelta],
              Union[datetime.datetime, datetime.timedelta]]:
    """This method gives you the most recently *completed* audit period.

    arguments:
            units: string, one of 'hour', 'day', 'month', 'year'
                    Periods normally begin at the beginning (UTC) of the
                    period unit (So a 'day' period begins at midnight UTC,
                    a 'month' unit on the 1st, a 'year' on Jan, 1)
                    unit string may be appended with an optional offset
                    like so:  'day@18'  This will begin the period at 18:00
                    UTC.  'month@15' starts a monthly period on the 15th,
                    and year@3 begins a yearly one on March 1st.


    returns:  2 tuple of datetimes (begin, end)
              The begin timestamp of this audit period is the same as the
              end of the previous.
    """
    if not unit:
        unit = CONF.volume_usage_audit_period

    unit = typing.cast(str, unit)

    offset: Union[str, int] = 0
    if '@' in unit:
        unit, offset = unit.split("@", 1)
        offset = int(offset)

    offset = typing.cast(int, offset)

    rightnow = timeutils.utcnow()
    if unit not in ('month', 'day', 'year', 'hour'):
        raise ValueError('Time period must be hour, day, month or year')
    if unit == 'month':
        if offset == 0:
            offset = 1
        end = datetime.datetime(day=offset,
                                month=rightnow.month,
                                year=rightnow.year)
        if end >= rightnow:
            year = rightnow.year
            if 1 >= rightnow.month:
                year -= 1
                month = 12 + (rightnow.month - 1)
            else:
                month = rightnow.month - 1
            end = datetime.datetime(day=offset,
                                    month=month,
                                    year=year)
        year = end.year
        if 1 >= end.month:
            year -= 1
            month = 12 + (end.month - 1)
        else:
            month = end.month - 1
        begin = datetime.datetime(day=offset, month=month, year=year)

    elif unit == 'year':
        if offset == 0:
            offset = 1
        end = datetime.datetime(day=1, month=offset, year=rightnow.year)
        if end >= rightnow:
            end = datetime.datetime(day=1,
                                    month=offset,
                                    year=rightnow.year - 1)
            begin = datetime.datetime(day=1,
                                      month=offset,
                                      year=rightnow.year - 2)
        else:
            begin = datetime.datetime(day=1,
                                      month=offset,
                                      year=rightnow.year - 1)

    elif unit == 'day':
        end = datetime.datetime(hour=offset,
                                day=rightnow.day,
                                month=rightnow.month,
                                year=rightnow.year)
        if end >= rightnow:
            end = end - datetime.timedelta(days=1)
        begin = end - datetime.timedelta(days=1)

    elif unit == 'hour':
        end = rightnow.replace(minute=offset, second=0, microsecond=0)
        if end >= rightnow:
            end = end - datetime.timedelta(hours=1)
        begin = end - datetime.timedelta(hours=1)

    return (begin, end)


def monkey_patch() -> None:
    """Patches decorators for all functions in a specified module.

    If the CONF.monkey_patch set as True,
    this function patches a decorator
    for all functions in specified modules.

    You can set decorators for each modules
    using CONF.monkey_patch_modules.
    The format is "Module path:Decorator function".
    Example: 'cinder.api.ec2.cloud:' \
     cinder.openstack.common.notifier.api.notify_decorator'

    Parameters of the decorator are as follows.
    (See cinder.openstack.common.notifier.api.notify_decorator)

    :param name: name of the function
    :param function: object of the function
    """
    # If CONF.monkey_patch is not True, this function do nothing.
    if not CONF.monkey_patch:
        return
    # Get list of modules and decorators
    for module_and_decorator in CONF.monkey_patch_modules:
        module, decorator_name = module_and_decorator.split(':')
        # import decorator function
        decorator = importutils.import_class(decorator_name)
        __import__(module)
        # Retrieve module information using pyclbr
        module_data = pyclbr.readmodule_ex(module)
        for key in module_data.keys():
            # set the decorator for the class methods
            if isinstance(module_data[key], pyclbr.Class):
                clz = importutils.import_class("%s.%s" % (module, key))
                # On Python 3, unbound methods are regular functions
                predicate = inspect.isfunction
                for method, func in inspect.getmembers(clz, predicate):
                    setattr(
                        clz, method,
                        decorator("%s.%s.%s" % (module, key, method), func))
            # set the decorator for the function
            elif isinstance(module_data[key], pyclbr.Function):
                func = importutils.import_class("%s.%s" % (module, key))
                setattr(sys.modules[module], key,
                        decorator("%s.%s" % (module, key), func))


def make_dev_path(dev: str,
                  partition: Optional[str] = None,
                  base: str = '/dev') -> str:
    """Return a path to a particular device.

    >>> make_dev_path('xvdc')
    /dev/xvdc

    >>> make_dev_path('xvdc', 1)
    /dev/xvdc1
    """
    path = os.path.join(base, dev)
    if partition:
        path += str(partition)
    return path


def robust_file_write(directory: str, filename: str, data: str) -> None:
    """Robust file write.

    Use "write to temp file and rename" model for writing the
    persistence file.

    :param directory: Target directory to create a file.
    :param filename: File name to store specified data.
    :param data: String data.
    """
    tempname = None
    dirfd = None
    try:
        dirfd = os.open(directory, os.O_DIRECTORY)

        # write data to temporary file
        with tempfile.NamedTemporaryFile(prefix=filename,
                                         dir=directory,
                                         delete=False) as tf:
            tempname = tf.name
            tf.write(data.encode('utf-8'))
            tf.flush()
            os.fdatasync(tf.fileno())
            tf.close()

            # Fsync the directory to ensure the fact of the existence of
            # the temp file hits the disk.
            os.fsync(dirfd)
            # If destination file exists, it will be replaced silently.
            os.rename(tempname, os.path.join(directory, filename))
            # Fsync the directory to ensure the rename hits the disk.
            os.fsync(dirfd)
    except OSError:
        with excutils.save_and_reraise_exception():
            LOG.error("Failed to write persistence file: %(path)s.",
                      {'path': os.path.join(directory, filename)})
            if tempname is not None:
                if os.path.isfile(tempname):
                    os.unlink(tempname)
    finally:
        if dirfd is not None:
            os.close(dirfd)


@contextlib.contextmanager
def temporary_chown(path: str,
                    owner_uid: Optional[int] = None) -> Iterator[None]:
    """Temporarily chown a path.

    :params owner_uid: UID of temporary owner (defaults to current user)
    """

    if os.name == 'nt':
        LOG.debug("Skipping chown for %s as this operation is "
                  "not available on Windows.", path)
        yield
        return

    if owner_uid is None:
        owner_uid = os.getuid()

    orig_uid = os.stat(path).st_uid

    if orig_uid != owner_uid:
        execute('chown', str(owner_uid), path, run_as_root=True)
    try:
        yield
    finally:
        if orig_uid != owner_uid:
            execute('chown', str(orig_uid), path, run_as_root=True)


@contextlib.contextmanager
def tempdir(**kwargs) -> Iterator[str]:
    tmpdir = tempfile.mkdtemp(**kwargs)
    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            LOG.debug('Could not remove tmpdir: %s', str(e))


def get_root_helper() -> str:
    return 'sudo cinder-rootwrap %s' % CONF.rootwrap_config


def get_file_mode(path: str) -> int:
    """This primarily exists to make unit testing easier."""
    return stat.S_IMODE(os.stat(path).st_mode)


def get_file_gid(path: str) -> int:
    """This primarily exists to make unit testing easier."""
    return os.stat(path).st_gid


def get_file_size(path: str) -> int:
    """Returns the file size."""
    return os.stat(path).st_size


def _get_disk_of_partition(
        devpath: str,
        st: Optional[os.stat_result] = None) -> tuple[str, os.stat_result]:
    """Gets a disk device path and status from partition path.

    Returns a disk device path from a partition device path, and stat for
    the device. If devpath is not a partition, devpath is returned as it is.
    For example, '/dev/sda' is returned for '/dev/sda1', and '/dev/disk1' is
    for '/dev/disk1p1' ('p' is prepended to the partition number if the disk
    name ends with numbers).
    """
    diskpath = re.sub(r'(?:(?<=\d)p)?\d+$', '', devpath)
    if diskpath != devpath:
        try:
            st_disk = os.stat(diskpath)
            if stat.S_ISBLK(st_disk.st_mode):
                return (diskpath, st_disk)
        except OSError:
            pass
    # devpath is not a partition
    if st is None:
        st = os.stat(devpath)
    return (devpath, st)


def get_bool_param(param_string: str,
                   params: dict,
                   default: bool = False) -> bool:
    param = params.get(param_string, default)
    if not strutils.is_valid_boolstr(param):
        msg = _("Value '%(param)s' for '%(param_string)s' is not "
                "a boolean.") % {'param': param, 'param_string': param_string}
        raise exception.InvalidParameterValue(err=msg)

    return strutils.bool_from_string(param, strict=True)


def get_blkdev_major_minor(path: str,
                           lookup_for_file: bool = True) -> Optional[str]:
    """Get 'major:minor' number of block device.

    Get the device's 'major:minor' number of a block device to control
    I/O ratelimit of the specified path.
    If lookup_for_file is True and the path is a regular file, lookup a disk
    device which the file lies on and returns the result for the device.
    """
    st = os.stat(path)
    if stat.S_ISBLK(st.st_mode):
        path, st = _get_disk_of_partition(path, st)
        return '%d:%d' % (os.major(st.st_rdev), os.minor(st.st_rdev))
    elif stat.S_ISCHR(st.st_mode):
        # No I/O ratelimit control is provided for character devices
        return None
    elif lookup_for_file:
        # lookup the mounted disk which the file lies on
        out, _err = execute('df', path)
        devpath = out.split("\n")[1].split()[0]
        if devpath[0] != '/':
            # the file is on a network file system
            return None
        return get_blkdev_major_minor(devpath, False)
    else:
        msg = _("Unable to get a block device for file \'%s\'") % path
        raise exception.CinderException(msg)


def check_string_length(value: str, name: str, min_length: int = 0,
                        max_length: Optional[int] = None,
                        allow_all_spaces: bool = True) -> None:
    """Check the length of specified string.

    :param value: the value of the string
    :param name: the name of the string
    :param min_length: the min_length of the string
    :param max_length: the max_length of the string
    """
    try:
        strutils.check_string_length(value, name=name,
                                     min_length=min_length,
                                     max_length=max_length)
    except (ValueError, TypeError) as exc:
        raise exception.InvalidInput(reason=exc)

    if not allow_all_spaces and value.isspace():
        msg = _('%(name)s cannot be all spaces.')
        raise exception.InvalidInput(reason=msg)


def is_blk_device(dev: str) -> bool:
    try:
        if stat.S_ISBLK(os.stat(dev).st_mode):
            return True
        return False
    except Exception:
        LOG.debug('Path %s not found in is_blk_device check', dev)
        return False


class ComparableMixin(object):
    def _compare(self, other: object, method: Callable):
        try:
            return method(self._cmpkey(), other._cmpkey())  # type: ignore
        except (AttributeError, TypeError):
            # _cmpkey not implemented, or return different type,
            # so I can't compare with "other".
            return NotImplemented

    def __lt__(self, other: object):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other: object):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other: object):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other: object):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other: object):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other: object):
        return self._compare(other, lambda s, o: s != o)


class retry_if_exit_code(tenacity.retry_if_exception):
    """Retry on ProcessExecutionError specific exit codes."""
    def __init__(self, codes):
        self.codes = (codes,) if isinstance(codes, int) else codes
        super(retry_if_exit_code, self).__init__(self._check_exit_code)

    def _check_exit_code(self, exc):
        return (exc and isinstance(exc, processutils.ProcessExecutionError) and
                exc.exit_code in self.codes)


def retry(retry_param: Union[None,
                             Type[Exception],
                             tuple[Type[Exception], ...],
                             int,
                             tuple[int, ...]],
          interval: float = 1,
          retries: int = 3,
          backoff_rate: float = 2,
          wait_random: bool = False,
          retry=tenacity.retry_if_exception_type) -> Callable:

    if retries < 1:
        raise ValueError('Retries must be greater than or '
                         'equal to 1 (received: %s). ' % retries)

    if wait_random:
        wait = tenacity.wait_random_exponential(multiplier=interval)
    else:
        wait = tenacity.wait_exponential(
            multiplier=interval, min=0, exp_base=backoff_rate)

    def _decorator(f: Callable) -> Callable:

        @functools.wraps(f)
        def _wrapper(*args, **kwargs):
            r = tenacity.Retrying(
                sleep=tenacity.nap.sleep,
                before_sleep=tenacity.before_sleep_log(LOG, logging.DEBUG),
                after=tenacity.after_log(LOG, logging.DEBUG),
                stop=tenacity.stop_after_attempt(retries),
                reraise=True,
                retry=retry(retry_param),
                wait=wait)
            return r(f, *args, **kwargs)

        return _wrapper

    return _decorator


def convert_str(text: Union[str, bytes]) -> str:
    """Convert to native string.

    Convert bytes and Unicode strings to native strings:

    * convert to bytes on Python 2:
      encode Unicode using encodeutils.safe_encode()
    * convert to Unicode on Python 3: decode bytes from UTF-8
    """
    if isinstance(text, bytes):
        return text.decode('utf-8')
    else:
        return text


def build_or_str(elements: Union[None, str, Iterable[str]],
                 str_format: Optional[str] = None) -> str:
    """Builds a string of elements joined by 'or'.

    Will join strings with the 'or' word and if a str_format is provided it
    will be used to format the resulted joined string.
    If there are no elements an empty string will be returned.

    :param elements: Elements we want to join.
    :type elements: String or iterable of strings.
    :param str_format: String to use to format the response.
    :type str_format: String.
    """
    if not elements:
        return ''

    if not isinstance(elements, str):
        elements = _(' or ').join(elements)

    elements = typing.cast(str, elements)

    if str_format:
        return str_format % elements

    return elements


def calculate_capacity_factors(total_capacity: float,
                               free_capacity: float,
                               provisioned_capacity: float,
                               thin_provisioning_support: bool,
                               max_over_subscription_ratio: float,
                               reserved_percentage: int,
                               thin: bool) -> dict:
    """Create the various capacity factors of the a particular backend.

    Based off of definition of terms
    cinder-specs/specs/queens/provisioning-improvements.html
    Description of factors calculated where units of gb are Gibibytes.
    reserved_capacity - The amount of space reserved from the total_capacity
    as reported by the backend.
    total_reserved_available_capacity - The total capacity minus reserved
    capacity
    total_available_capacity - The total capacity available to cinder
    calculated from total_reserved_available_capacity (for thick) OR
    for thin total_reserved_available_capacity max_over_subscription_ratio
    calculated_free_capacity - total_available_capacity - provisioned_capacity
    virtual_free_capacity - The calculated free capacity available to cinder
    to allocate new storage.
    For thin: calculated_free_capacity
    For thick: the reported free_capacity can be less than the calculated
    capacity, so we use free_capacity - reserved_capacity.

    free_percent - the percentage of the virtual_free_capacity and
    total_available_capacity is left over
    provisioned_ratio - The ratio of provisioned storage to
    total_available_capacity

    :param total_capacity: The reported total capacity in the backend.
    :type total_capacity: float
    :param free_capacity: The free space/capacity as reported by the backend.
    :type free_capacity: float
    :param provisioned_capacity: as reported by backend or volume manager from
        allocated_capacity_gb
    :type provisioned_capacity: float
    :param thin_provisioning_support: Is thin provisioning supported?
    :type thin_provisioning_support: bool
    :param max_over_subscription_ratio: as reported by the backend
    :type max_over_subscription_ratio: float
    :param reserved_percentage: the % amount to reserve as unavailable. 0-100
    :type reserved_percentage: int, 0-100
    :param thin: calculate based on thin provisioning if enabled by
        thin_provisioning_support
    :type thin: bool
    :return: A dictionary of all of the capacity factors.
    :rtype: dict

    """

    total = float(total_capacity)
    reserved = float(reserved_percentage) / 100
    reserved_capacity = math.floor(total * reserved)
    total_reserved_available = total - reserved_capacity

    if thin and thin_provisioning_support:
        total_available_capacity = (
            total_reserved_available * max_over_subscription_ratio
        )
        calculated_free = total_available_capacity - provisioned_capacity
        virtual_free = calculated_free
        provisioned_type = 'thin'
    else:
        # Calculate how much free space is left after taking into
        # account the reserved space.
        total_available_capacity = total_reserved_available
        calculated_free = total_available_capacity - provisioned_capacity
        virtual_free = calculated_free
        if free_capacity < calculated_free:
            virtual_free = free_capacity

        provisioned_type = 'thick'

    if total_available_capacity:
        provisioned_ratio = provisioned_capacity / total_available_capacity
        free_percent = (virtual_free / total_available_capacity) * 100
    else:
        provisioned_ratio = 0
        free_percent = 0

    def _limit(x):
        """Limit our floating points to 2 decimal places."""
        return round(x, 2)

    return {
        "total_capacity": total,
        "free_capacity": free_capacity,
        "reserved_capacity": reserved_capacity,
        "total_reserved_available_capacity": _limit(total_reserved_available),
        "max_over_subscription_ratio": (
            max_over_subscription_ratio if provisioned_type == 'thin' else None
        ),
        "total_available_capacity": _limit(total_available_capacity),
        "provisioned_capacity": provisioned_capacity,
        "calculated_free_capacity": _limit(calculated_free),
        "virtual_free_capacity": _limit(virtual_free),
        "free_percent": _limit(free_percent),
        "provisioned_ratio": _limit(provisioned_ratio),
        "provisioned_type": provisioned_type
    }


def calculate_virtual_free_capacity(total_capacity: float,
                                    free_capacity: float,
                                    provisioned_capacity: float,
                                    thin_provisioning_support: bool,
                                    max_over_subscription_ratio: float,
                                    reserved_percentage: int,
                                    thin: bool) -> float:
    """Calculate the virtual free capacity based on multiple factors.

    :param total_capacity:  total_capacity_gb of a host_state or pool.
    :param free_capacity:   free_capacity_gb of a host_state or pool.
    :param provisioned_capacity:    provisioned_capacity_gb of a host_state
                                    or pool.
    :param thin_provisioning_support:   thin_provisioning_support of
                                        a host_state or a pool.
    :param max_over_subscription_ratio: max_over_subscription_ratio of
                                        a host_state or a pool
    :param reserved_percentage: reserved_percentage of a host_state or
                                a pool.
    :param thin: whether volume to be provisioned is thin
    :returns: the calculated virtual free capacity.
    """

    factors = calculate_capacity_factors(
        total_capacity,
        free_capacity,
        provisioned_capacity,
        thin_provisioning_support,
        max_over_subscription_ratio,
        reserved_percentage,
        thin
    )
    return factors["virtual_free_capacity"]


def calculate_max_over_subscription_ratio(
        capability: dict,
        global_max_over_subscription_ratio: float) -> float:
    # provisioned_capacity_gb is the apparent total capacity of
    # all the volumes created on a backend, which is greater than
    # or equal to allocated_capacity_gb, which is the apparent
    # total capacity of all the volumes created on a backend
    # in Cinder. Using allocated_capacity_gb as the default of
    # provisioned_capacity_gb if it is not set.
    allocated_capacity_gb = capability.get('allocated_capacity_gb', 0)
    provisioned_capacity_gb = capability.get('provisioned_capacity_gb',
                                             allocated_capacity_gb)
    thin_provisioning_support = capability.get('thin_provisioning_support',
                                               False)
    total_capacity_gb = capability.get('total_capacity_gb', 0)
    free_capacity_gb = capability.get('free_capacity_gb', 0)
    pool_name = capability.get('pool_name',
                               capability.get('volume_backend_name'))

    # If thin provisioning is not supported the capacity filter will not use
    # the value we return, no matter what it is.
    if not thin_provisioning_support:
        LOG.debug("Trying to retrieve max_over_subscription_ratio from a "
                  "service that does not support thin provisioning")
        return 1.0

    # Again, if total or free capacity is infinite or unknown, the capacity
    # filter will not use the max_over_subscription_ratio at all. So, does
    # not matter what we return here.
    if ((total_capacity_gb in INFINITE_UNKNOWN_VALUES) or
            (free_capacity_gb in INFINITE_UNKNOWN_VALUES)):
        return 1.0

    max_over_subscription_ratio = (capability.get(
        'max_over_subscription_ratio') or global_max_over_subscription_ratio)

    # We only calculate the automatic max_over_subscription_ratio (mosr)
    # when the global or driver conf is set auto and while
    # provisioned_capacity_gb is not 0. When auto is set and
    # provisioned_capacity_gb is 0, we use the default value 20.0.
    if max_over_subscription_ratio == 'auto':
        if provisioned_capacity_gb != 0:
            used_capacity = total_capacity_gb - free_capacity_gb
            LOG.debug("Calculating max_over_subscription_ratio for "
                      "pool %s: provisioned_capacity_gb=%s, "
                      "used_capacity=%s",
                      pool_name, provisioned_capacity_gb, used_capacity)
            max_over_subscription_ratio = 1 + (
                float(provisioned_capacity_gb) / (used_capacity + 1))
        else:
            max_over_subscription_ratio = INITIAL_AUTO_MOSR

        LOG.info("Auto max_over_subscription_ratio for pool %s is "
                 "%s", pool_name, max_over_subscription_ratio)
    else:
        max_over_subscription_ratio = float(max_over_subscription_ratio)

    return max_over_subscription_ratio


def validate_dictionary_string_length(specs: dict) -> None:
    """Check the length of each key and value of dictionary."""
    if not isinstance(specs, dict):
        msg = _('specs must be a dictionary.')
        raise exception.InvalidInput(reason=msg)

    for key, value in specs.items():
        if key is not None:
            check_string_length(key, 'Key "%s"' % key,
                                min_length=1, max_length=255)

        if value is not None:
            check_string_length(value, 'Value for key "%s"' % key,
                                min_length=0, max_length=255)


def service_expired_time(
        with_timezone: Optional[bool] = False) -> datetime.datetime:
    return (timeutils.utcnow(with_timezone=with_timezone) -
            datetime.timedelta(seconds=CONF.service_down_time))


class DoNothing(str):
    """Class that literrally does nothing.

    We inherit from str in case it's called with json.dumps.
    """
    def __call__(self, *args, **kwargs):
        return self

    def __getattr__(self, name):
        return self


DO_NOTHING = DoNothing()


def notifications_enabled(conf):
    """Check if oslo notifications are enabled."""
    notifications_driver = set(conf.oslo_messaging_notifications.driver)
    return notifications_driver and notifications_driver != {'noop'}


def if_notifications_enabled(f: Callable) -> Callable:
    """Calls decorated method only if notifications are enabled."""
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if notifications_enabled(CONF):
            return f(*args, **kwargs)
        return DO_NOTHING
    return wrapped


LOG_LEVELS = ('INFO', 'WARNING', 'ERROR', 'DEBUG')


def get_log_method(level_string: str) -> int:
    level_string = level_string or ''
    upper_level_string = level_string.upper()
    if upper_level_string not in LOG_LEVELS:
        raise exception.InvalidInput(
            reason=_('%s is not a valid log level.') % level_string)
    return getattr(logging, upper_level_string)


def set_log_levels(prefix: str, level_string: str) -> None:
    level = get_log_method(level_string)
    prefix = prefix or ''

    for k, v in logging.get_loggers().items():
        if k and k.startswith(prefix):
            v.logger.setLevel(level)


def get_log_levels(prefix: str) -> dict:
    prefix = prefix or ''
    return {k: py_logging.getLevelName(v.logger.getEffectiveLevel())
            for k, v in logging.get_loggers().items()
            if k and k.startswith(prefix)}


def paths_normcase_equal(path_a: str, path_b: str) -> bool:
    return os.path.normcase(path_a) == os.path.normcase(path_b)


def create_ordereddict(adict: dict) -> OrderedDict:
    """Given a dict, return a sorted OrderedDict."""
    return OrderedDict(sorted(adict.items(),
                              key=operator.itemgetter(0)))


class Semaphore(object):
    """Custom semaphore to workaround eventlet issues with multiprocessing."""
    def __init__(self, limit):
        self.limit = limit
        self.semaphore = multiprocessing.Semaphore(limit)

    def __enter__(self):
        # Eventlet does not work with multiprocessing's Semaphore, so we have
        # to execute it in a native thread to avoid getting blocked when trying
        # to acquire the semaphore.
        return tpool.execute(self.semaphore.__enter__)

    def __exit__(self, *args):
        # Don't use native thread for exit, as it will only add overhead
        return self.semaphore.__exit__(*args)


def semaphore_factory(limit: int,
                      concurrent_processes: int) -> Union[eventlet.Semaphore,
                                                          Semaphore]:
    """Get a semaphore to limit concurrent operations.

    The semaphore depends on the limit we want to set and the concurrent
    processes that need to be limited.
    """
    # Limit of 0 is no limit, so we won't use a semaphore
    if limit:
        # If we only have 1 process we can use eventlet's Semaphore
        if concurrent_processes == 1:
            return eventlet.Semaphore(limit)
        # Use our own Sempahore for interprocess because eventlet blocks with
        # the standard one
        return Semaphore(limit)
    return contextlib.suppress()


def limit_operations(func: Callable) -> Callable:
    """Decorator to limit the number of concurrent operations.

     This method decorator expects to have a _semaphore attribute holding an
     initialized semaphore in the self instance object.

     We can get the appropriate semaphore with the semaphore_factory method.
     """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        with self._semaphore:
            return func(self, *args, **kwargs)
    return wrapper
